#include <glib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct _Summary Summary;
typedef struct _SummaryBlock SummaryBlock;
typedef struct _SummaryItem SummaryItem;

struct _SummaryItem {
	char *uid;
	int seq;
	int flags;
	size_t size;

	union {
		struct {
			off_t from_offset;
			off_t subject_offset;
			off_t to_offset;
			off_t cc_offset;
		} offsets;
		struct {
			char *from;
			char *subject;
			char *to;
			char *cc;
		} pointers;
	} mmappable;

	int mem;
	SummaryBlock *block;
};

struct _SummaryBlock {
	int ref_count;
	char *data;
	FILE *data_file;
	int mmap_size;
	int block_id;
	int written;
	GHashTable *uids;
	GHashTable *seqs;
	GStaticRecMutex *lock;
};

struct _Summary {
	GPtrArray *blocks;
	GStaticRecMutex *lock;
};

static char*
summary_block_index_filename (SummaryBlock *b)
{
	return g_strdup_printf ("index_%d.idx", b->block_id);
}

static char*
summary_block_flags_filename (SummaryBlock *b)
{
	return g_strdup_printf ("flags_%d.idx", b->block_id);
}

static char*
summary_block_data_filename (SummaryBlock *b)
{
	return g_strdup_printf ("data_%d.mmap", b->block_id);
}

static void 
prepare_summary_block (SummaryBlock *b)
{
	b->uids = g_hash_table_new_full (g_str_hash, g_str_equal, (GDestroyNotify) g_free, NULL);
	b->seqs = g_hash_table_new_full (g_int_hash, g_int_equal, NULL, NULL);
}

static SummaryBlock*
create_new_summary_block (int block_id)
{
	SummaryBlock *b = g_slice_new0 (SummaryBlock);
	b->lock = g_new0 (GStaticRecMutex, 1);
	g_static_rec_mutex_init (b->lock);

	g_static_rec_mutex_lock (b->lock);
	b->block_id = block_id;
	b->written = FALSE;
	prepare_summary_block (b);
	g_static_rec_mutex_unlock (b->lock);

	return b;
}

static void 
summary_block_add_item (SummaryBlock *b, SummaryItem *i)
{
	g_static_rec_mutex_lock (b->lock);
	b->ref_count++;
	b->written = FALSE;
	g_hash_table_insert (b->uids, g_strdup (i->uid), i);
	g_hash_table_insert (b->seqs, &i->seq, i);
	g_static_rec_mutex_unlock (b->lock);
}

void
summary_add_item (Summary *s, SummaryItem *i)
{
	int block_id = 0; //i->seq % 1000;
	SummaryBlock *b;

	g_static_rec_mutex_lock (s->lock);

	if (s->blocks->len < block_id+1) {
		int t;
		for (t = s->blocks->len; t < block_id+1; t++) {
			SummaryBlock *mnew = create_new_summary_block (t);
			g_ptr_array_add (s->blocks, mnew);
		}
	}

	b = (SummaryBlock *) s->blocks->pdata[block_id];
	summary_block_add_item (b, i);

	g_static_rec_mutex_unlock (s->lock);

}

static SummaryItem*
summary_block_get_item_by_seq (SummaryBlock *b, int seq)
{
	return (SummaryItem *) g_hash_table_lookup (b->seqs, &seq);
}

SummaryItem*
summary_get_item_by_seq (Summary *s, int seq)
{
	int i;
	SummaryItem *item = NULL;

	g_static_rec_mutex_lock (s->lock);

	for (i = 0; i < s->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) s->blocks->pdata[i];
		item = summary_block_get_item_by_seq (b, seq);
		if (item) {
			item->block->ref_count++;
			break;
		}
	}

	g_static_rec_mutex_unlock (s->lock);

	return item;
}

static SummaryItem*
summary_block_get_item_by_uid (SummaryBlock *b, char *uid)
{
	return (SummaryItem *) g_hash_table_lookup (b->uids, uid);
}

SummaryItem*
summary_get_item_by_uid (Summary *s, char *uid)
{
	int i;
	SummaryItem *item = NULL;

	g_static_rec_mutex_lock (s->lock);

	for (i = 0; i < s->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) s->blocks->pdata[i];
		item = summary_block_get_item_by_uid (b, uid);
		if (item) {
			item->block->ref_count++;
			break;
		}
	}

	g_static_rec_mutex_unlock (s->lock);

	return item;
}


static SummaryBlock *
read_summary_block (int block_id, SummaryBlock *b_in)
{
	FILE *idx, *flags;
	SummaryBlock *b;
	struct stat buf;
	char *index_filename;
	char *data_filename;
	char *flags_filename;

	if (b_in)
		b = b_in;
	else {
		b = g_slice_new0 (SummaryBlock);
		b->lock = g_new0 (GStaticRecMutex, 1);
		g_static_rec_mutex_init (b->lock);
	}

	g_static_rec_mutex_lock (b->lock);

	b->block_id = block_id;
	b->written = TRUE;

	prepare_summary_block (b);

	index_filename = summary_block_index_filename (b);

	idx = fopen (index_filename, "r");

	if (!idx) {
		b->written = FALSE;
		g_static_rec_mutex_unlock (b->lock);
		return b;
	}

	data_filename = summary_block_data_filename (b);

	if (b->data_file)
		fclose (b->data_file);
	if (b->data)
		munmap (b->data, b->mmap_size);

	b->data_file = fopen (data_filename, "r");

	if (!b->data_file) {
		b->written = FALSE;
		fclose (idx);
		g_free (index_filename);
		g_free (data_filename);
		g_static_rec_mutex_unlock (b->lock);
		return b;
	}

	flags_filename = summary_block_flags_filename (b);
	flags = fopen (flags_filename, "r");
	if (!flags) {
		b->written = FALSE;
		fclose (idx);
		fclose (b->data_file);
		b->data_file = NULL;
		g_free (index_filename);
		g_free (data_filename);
		g_free (flags_filename);
		g_static_rec_mutex_unlock (b->lock);
		return b;
	}

	g_free (index_filename);
	g_free (data_filename);
	g_free (flags_filename);

	fstat(fileno (b->data_file), &buf);
	b->mmap_size = buf.st_size;
	b->data = mmap (NULL, b->mmap_size, PROT_READ, MAP_PRIVATE, fileno (b->data_file), 0);

	while (!feof (idx)) {
		int len;
		if (fscanf (idx, "%d", &len) != EOF) 
		{
			SummaryItem *cur;
			char *uid = (char *) malloc(len);

			memset (uid, 0, len);
			fgetc (idx);
			fread (uid, 1, len, idx);
			uid[len] = '\0';

			cur = summary_block_get_item_by_uid (b, uid);
			if (!cur)
				cur = g_slice_new0 (SummaryItem);
			cur->uid = uid;
			cur->block = b;
			cur->block->ref_count++;

			if (cur->mem) {
				g_free (cur->mmappable.pointers.from);
				g_free (cur->mmappable.pointers.subject);
				g_free (cur->mmappable.pointers.to);
				g_free (cur->mmappable.pointers.cc);
			}
			cur->mem = 0;

			fscanf (idx, "%d%d%d%d%d%d",
				&cur->seq, &cur->size,
				(int *) &cur->mmappable.offsets.subject_offset, 
				(int *) &cur->mmappable.offsets.from_offset,
				(int *) &cur->mmappable.offsets.to_offset,
				(int *) &cur->mmappable.offsets.cc_offset);

			summary_block_add_item (b, cur);
		}
	}	
	fclose (idx);

	/* Perhaps scan this in a table and loop the table in above loop  in stead */
	while (!feof (flags)) {
		int seq, flag;
		if (fscanf (flags, "%d%d", &seq, &flag) != EOF) {
			SummaryItem *c = summary_block_get_item_by_seq (b, seq);
			if (c)
				c->flags = flag;
		}
	}
	fclose(flags);

	g_static_rec_mutex_unlock (b->lock);

	return b;
}

const char*
summary_item_get_uid (SummaryItem *i)
{
	return i->uid;
}

int
summary_item_get_seq (SummaryItem *i)
{
	return (const int) i->seq;
}

int
summary_item_get_flags (SummaryItem *i)
{
	return (const int) i->flags;
}

void
summary_item_set_flags (SummaryItem *i, int new_flags)
{
	i->flags = new_flags;
	/* todo: rewrite flags file */
	return;
}


const char*
summary_item_get_subject (SummaryItem *i)
{
	const char* retval;

	if (i->block)
		g_static_rec_mutex_lock (i->block->lock);
	if (i->mem)
		retval = i->mmappable.pointers.subject;
	else
		retval = i->block->data + i->mmappable.offsets.subject_offset;
	if (i->block)
		g_static_rec_mutex_unlock (i->block->lock);

	return retval;
}

const char*
summary_item_get_from (SummaryItem *i)
{
	const char *retval;

	if (i->block)
		g_static_rec_mutex_lock (i->block->lock);
	if (i->mem)
		retval = i->mmappable.pointers.from;
	else 
		retval = i->block->data + i->mmappable.offsets.from_offset;
	if (i->block)
		g_static_rec_mutex_unlock (i->block->lock);

	return retval;
}


const char*
summary_item_get_to (SummaryItem *i)
{
	const char *retval;

	if (i->block)
		g_static_rec_mutex_lock (i->block->lock);
	if (i->mem)
		retval = i->mmappable.pointers.to;
	else
		retval = i->block->data + i->mmappable.offsets.to_offset;
	if (i->block)
		g_static_rec_mutex_unlock (i->block->lock);

	return retval;
}

const char*
summary_item_get_cc (SummaryItem *i)
{
	const char *retval;

	if (i->block)
		g_static_rec_mutex_lock (i->block->lock);
	if (i->mem)
		retval = i->mmappable.pointers.cc;
 	else
		retval = i->block->data + i->mmappable.offsets.cc_offset;
	if (i->block)
		g_static_rec_mutex_unlock (i->block->lock);

	return retval;
}

static void 
summary_block_kill (SummaryBlock *b)
{
	if (b->data)
		munmap (b->data, b->mmap_size);
	if (b->data_file)
		fclose (b->data_file);

	g_hash_table_destroy (b->seqs);
	g_hash_table_destroy (b->uids);

	g_free (b->lock);
	g_slice_free (SummaryBlock, b);

}


static void
summary_item_free (SummaryItem *i)
{
	if (i->block) {
		g_static_rec_mutex_lock (i->block->lock);
		i->block->ref_count--;
		if (i->block->ref_count <= 0)
			summary_block_kill (i->block);
		g_static_rec_mutex_unlock (i->block->lock);
	}

	g_free (i->uid);
	if (i->mem) {
		g_free (i->mmappable.pointers.from);
		g_free (i->mmappable.pointers.subject);
		g_free (i->mmappable.pointers.to);
		g_free (i->mmappable.pointers.cc);
		i->mem = 0;
	}
	g_slice_free (SummaryItem, i);
}


typedef struct {
	char *str;
	int cnt, cpy;
	GList *affected;
} StoreString;


static gint 
cmp_storestring (gconstpointer a, gconstpointer b)
{
	StoreString *sa = (StoreString *) a;
	StoreString *sb = (StoreString *) b;
	return sb->cnt - sa->cnt;
}


static void
summary_block_dump (SummaryBlock *b)
{
	FILE *data, *idx, *flags;
	GList *orig = g_hash_table_get_values (b->seqs), *items = orig;
	GHashTable *strings = g_hash_table_new (g_str_hash, g_str_equal);
	GList *sstrings;

	char *index_filename;
	char *data_filename;
	char *flags_filename;
	char *index_filename_n;
	char *data_filename_n;
	char *flags_filename_n;

	g_static_rec_mutex_lock (b->lock);

	if (b->written) {
		g_static_rec_mutex_unlock (b->lock);
		return;
	}

	index_filename = summary_block_index_filename (b);
	data_filename = summary_block_data_filename (b);
	flags_filename = summary_block_flags_filename (b);

	index_filename_n = g_strdup_printf ("%s~", index_filename);
	data_filename_n = g_strdup_printf ("%s~", data_filename);
	flags_filename_n = g_strdup_printf ("%s~", flags_filename);

	idx = fopen (index_filename_n, "w");
	data = fopen (data_filename_n, "w");
	flags = fopen (flags_filename_n, "w");


	while (items) {
		SummaryItem *item = (SummaryItem *) items->data;

		StoreString *sb;
		char *str;

		/* subject */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.subject_offset;
		else
			str = item->mmappable.pointers.subject;

		sb =  g_hash_table_lookup (strings, str);
		if (sb) {
			sb->cnt++;
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = item->mem ? g_strdup (item->mmappable.pointers.subject) : str;
			sb->cpy = item->mem;
			g_hash_table_insert (strings, str, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.subject_offset);

		/* from */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.from_offset;
		else
			str = item->mmappable.pointers.from;

		sb = g_hash_table_lookup (strings, str);
		if (sb) {
			sb->cnt++;
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = item->mem ? g_strdup (item->mmappable.pointers.from) : str;
			sb->cpy = item->mem;
			g_hash_table_insert (strings, str, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.from_offset);


		/* to */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.to_offset;
		else
			str = item->mmappable.pointers.to;

		sb = g_hash_table_lookup (strings, str);
		if (sb) {
			sb->cnt++;
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = item->mem ? g_strdup (item->mmappable.pointers.to) : str;
			sb->cpy = item->mem;
			g_hash_table_insert (strings, str, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.to_offset);


		/* cc */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.cc_offset;
		else
			str = item->mmappable.pointers.cc;

		sb = g_hash_table_lookup (strings, str);
		if (sb) {
			sb->cnt++;
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = item->mem ? g_strdup (item->mmappable.pointers.cc) : str;
			sb->cpy = item->mem;
			g_hash_table_insert (strings, str, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.cc_offset);

		if (item->mem) {
			g_free (item->mmappable.pointers.from);
			g_free (item->mmappable.pointers.subject);
			g_free (item->mmappable.pointers.to);
			g_free (item->mmappable.pointers.cc);
			item->mem = 0;
		}

		items = g_list_next (items);
	}

	sstrings = g_hash_table_get_values (strings);
	sstrings = g_list_sort (sstrings, cmp_storestring);

	while (sstrings) {
		StoreString *ss = (StoreString *) sstrings->data;
		int location = ftell (data);

		fputs (ss->str, data);
		if (ss->cpy) {
			g_free (ss->str);
			ss->cpy = FALSE;
		}
		fputc ('\0', data);

		while (ss->affected) {
			int *data = ss->affected->data;
			*data = location;
			ss->affected = g_list_next (ss->affected);
		}
		g_list_free (ss->affected);
		
		g_slice_free (StoreString, ss);
		sstrings = g_list_next (sstrings);
	}

	g_list_free (sstrings);
	g_hash_table_destroy (strings);

	fclose (data);

	items = orig;
	while (items) {
		SummaryItem *item = (SummaryItem *) items->data;

		fprintf (flags, "%d %d\n", item->seq, item->flags); 
		fprintf (idx, "%d %s %d %d %d %d %d %d\n",
				strlen (item->uid), item->uid,
				item->seq, item->size,
				(int) item->mmappable.offsets.subject_offset, 
				(int) item->mmappable.offsets.from_offset,
				(int) item->mmappable.offsets.to_offset,
				(int) item->mmappable.offsets.cc_offset);

		items = g_list_next (items);
	}


	g_list_free (orig);

	fclose (idx);
	fclose (flags);

	if (b->data)
		munmap (b->data, b->mmap_size);
	if (b->data_file)
		fclose (b->data_file);

	b->data = NULL;
	b->data_file = NULL;

	rename (index_filename_n, index_filename);
	rename (data_filename_n, data_filename);
	rename (flags_filename_n, flags_filename);

	read_summary_block (b->block_id, b);

	g_free (index_filename);
	g_free (data_filename);
	g_free (flags_filename);
	g_free (index_filename_n);
	g_free (data_filename_n);
	g_free (flags_filename_n);

	g_static_rec_mutex_unlock (b->lock);

}

static void
summary_dump (Summary *summary)
{
	int i;
	for (i = 0; i < summary->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
		summary_block_dump (b);
	}
}

static void
foreach_item_free (gpointer key, gpointer value, gpointer user_data)
{
	SummaryItem *i = (SummaryItem *) value;
	summary_item_free (i);
}


static void
summary_block_free (SummaryBlock *b)
{
	g_static_rec_mutex_lock (b->lock);

	g_hash_table_foreach (b->seqs, foreach_item_free, NULL);

	if (b->ref_count <= 0) {
		g_static_rec_mutex_unlock (b->lock);
		summary_block_kill (b);
	} else
		g_static_rec_mutex_unlock (b->lock);
}

static void
summary_free (Summary *summary)
{
	int i;
	g_static_rec_mutex_lock (summary->lock);

	for (i = 0; i < summary->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
		summary_block_free (b);
	}
	g_ptr_array_free (summary->blocks, TRUE);

	g_static_rec_mutex_unlock (summary->lock);

	g_free (summary->lock);
	g_slice_free (Summary, summary);
}

static Summary*
read_summary (void)
{
	Summary *summary = g_slice_new0 (Summary);
	summary->lock = g_new0 (GStaticRecMutex, 1);
	int i, items;
	FILE *f = fopen ("summary.idx", "r");

	g_static_rec_mutex_init (summary->lock);

	fscanf (f, "%d", &items);

	g_static_rec_mutex_lock (summary->lock);

	summary->blocks = g_ptr_array_sized_new (items);

	for (i=0; i < items; i++) {
		SummaryBlock *b = read_summary_block (i, NULL);
		g_ptr_array_add (summary->blocks, b);
	}

	g_static_rec_mutex_unlock (summary->lock);

	return summary;
}

SummaryItem*
create_summary_item (const char *uid, int seq, int flags, size_t size, const char *from, const char *subject, const char *to, const char *cc)
{
	SummaryItem *item = g_slice_new0 (SummaryItem);

	item->size = size;
	item->seq = seq;
	item->flags = flags;
	item->uid = g_strdup (uid);
	item->mmappable.pointers.from = g_strdup (from);
	item->mmappable.pointers.to = g_strdup (to);
	item->mmappable.pointers.cc = g_strdup (cc);
	item->mmappable.pointers.subject = g_strdup (subject);
	item->block = NULL;
	item->mem = 1; /* memory item */

	return item;
}

static void 
foreach_item_print_it (gpointer key, gpointer value, gpointer user_data)
{
	SummaryItem *mc = (SummaryItem *) value;

	printf ("UID %s is at sequence %d with flags %d\n", 
		summary_item_get_uid (mc),
		summary_item_get_seq (mc),
		summary_item_get_flags (mc));

	printf ("\tFrom=%s\n\tSubj=%s\n\tCc=%s\n\tTo=%s\n",
		summary_item_get_from (mc),
		summary_item_get_subject (mc),
		summary_item_get_cc (mc),
		summary_item_get_to (mc));
}

static void 
print_summary (Summary *summary)
{
	int i;

	for (i = 0; i < summary->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
		g_hash_table_foreach (b->seqs, foreach_item_print_it, NULL);
	}
}

int main (int argc, char **argv)
{
	Summary *summary = read_summary ();
	int i, y = 0;

	print_summary (summary);

	if (argc > 1) {

		for (i = 0; i < summary->blocks->len; i++) {
			SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
			y += g_hash_table_size (b->seqs);
		}

		for (i = y; i < y + 50000; i++) {
			const char *uid = g_strdup_printf ("uid%d", i);
			int seq = 10+i;
			int flags = i+1;
			size_t size = i+2;
			char *from = g_strdup_printf ("from%d", i);
			const char *subject = g_strdup_printf ("subj%d", i);
			const char *to = g_strdup_printf ("to%d", i);
			const char *cc = g_strdup_printf ("cc%d", i);

			SummaryItem *item = create_summary_item (uid, seq, flags, size, from, subject, to, cc);
		
			summary_add_item (summary, item);
		}

		summary_dump (summary);

		printf ("\n\n\n\n");

		print_summary (summary);

	}

	summary_free (summary);

	return 0;
}

